#include "Application.h"
// Prototypes
bool deleteNode(void *n);
int compareNames(const void* first, const void *second);
bool writeToFile(void *data, void *f);

// Node
Node::Node(BString name, BMessage *archive) : name(name), archive(archive) {}

// FieldList
bool deleteNode(void *n) {
Node *node = (Node*)n;
	if (node->archive != NULL) delete node->archive;
	delete node;
	return false;
}

FieldList::~FieldList() {
	list.DoForEach(deleteNode);
}

BMessage *FieldList::ItemAt(int32 i){
Node *node = (Node*)list.ItemAt(i);
	if (node != NULL)
		return node->archive;
	else
		return NULL;
}

void FieldList::Remove(int32 i) {
	list.RemoveItem(i);
}

bool writeToFile(void *data, void *f) {
Node *node = (Node*) data;
BFile *file = (BFile*)f;
int32 size = node->name.Length()+1;
	file->Write(&size, sizeof(size));
	file->Write(node->name.LockBuffer(size+1), size);
	node->name.UnlockBuffer();
	node->archive->Flatten(file);	
	return false;
}

void FieldList::Save(BFile *file) {
int32 size = list.CountItems();
	file->Write(&size, sizeof(size));
	list.DoForEach(writeToFile, file);
}

int compareNames(const void* first, const void *second) {
	return ((Node*)*((const void**)first))->name.Compare(((Node*)*((const void**)second))->name);
}

void FieldList::Restore(BFile *file) {
int32 size, length;
char string[80];
BMessage *archive;
	file->Read(&size, sizeof(size));
	list = BList();
	for (;size; size--) {
		file->Read(&length, sizeof(length));
		file->Read(string, length);
		archive = new BMessage();
		archive->Unflatten(file);		
		list.AddItem(new Node(BString(string), archive));
	}
	list.SortItems(compareNames);
}

void FieldList::BuildMenu(BMenu *menu, int32 id){
int32 l = list.CountItems(), i;
Node *node;
	for (i = menu->CountItems(); i; i--) menu->RemoveItem((int32)0);
	
	for (i = 0; i < l; i++) {
		node = (Node*)list.ItemAt(i);
		BMessage *msg = new BMessage(id);
		msg->AddInt32("index", i);
		menu->AddItem(new BMenuItem(node->name.LockBuffer(node->name.Length()+1), msg));
		node->name.UnlockBuffer();
	}
}

int32 FieldList::Find(BString name){
int32 l = list.CountItems();
Node *node;
	for (int32 i = 0; i < l; i++) {
		node = (Node*)list.ItemAt(i);
		if (name.Compare(node->name) == 0) return i;
	}
	return -1;
}

void FieldList::Replace(int32 i, BMessage *archive){
Node *node = (Node*)list.ItemAt(i);
	if (node != NULL) {
		if (node->archive != NULL) delete node->archive;
		node->archive = archive;
	}
}

void FieldList::Insert(BString name, BMessage *archive){
	list.AddItem(new Node(name, archive));
	list.SortItems(compareNames);
}

// AppWindow
AppWindow::AppWindow(BRect aRect) 
	: BWindow(aRect, "Solitaire", B_TITLED_WINDOW, B_OUTLINE_RESIZE) {
	// add menu bar
	BRect rect = BRect(0,0,aRect.Width(), aRect.Height());

	app_info info;
	if (B_OK == be_app->GetAppInfo(&info)) {
		BEntry entry;
		entry = BEntry(&info.ref);
		entry.GetPath(&appPath);
		appPath.GetParent(&appPath);
	} 

	RestorePreferences();
	menubar = new BMenuBar(rect, "menu_bar");
	BMenu *menu; 
	BMenuItem *item; 

	menu = new BMenu("Game");
	menu->AddItem(new BMenuItem("New", new BMessage(MENU_APP_NEW), 'N'));
	menu->AddItem(new BMenuItem("Undo", new BMessage(MENU_GAME_UNDO), 'U'));
	menu->AddItem(new BMenuItem("Redo", new BMessage(MENU_GAME_REDO), 'R'));
	menu->AddItem(new BMenuItem("Edit", new BMessage(MENU_GAME_EDIT), 'E'));
	menu->AddItem(item = new BMenuItem("Save", new BMessage(MENU_GAME_SAVE), 'S'));
	item->SetEnabled(false);
	menu->AddItem(item = new BMenuItem("Delete", new BMessage(MENU_GAME_DELETE), 'D'));
	item->SetEnabled(false);
	menu->AddSeparatorItem();
	menu->AddItem(new BMenuItem("About ...", new BMessage(B_ABOUT_REQUESTED)));
	menu->AddSeparatorItem();
	menu->AddItem(new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED), 'Q')); 
	menubar->AddItem(menu);

	field = menu = new BMenu("Field");
	menu->SetRadioMode(true);
	fieldList.BuildMenu(menu, MENU_FIELD);
	if (menu->CountItems() > 0) 
		menu->ItemAt(0)->SetMarked(true);
	menubar->AddItem(menu);

	AddChild(menubar);
	// add view
	aRect.Set(0, menubar->Bounds().Height()+1, aRect.Width(), aRect.Height());
	view = NULL;
	AddChild(view = new View(aRect));
	UpdateUnRedoMenu();
	// make window visible
	Show();
}

// SaveWindow

class SaveWindow : public BWindow {
	BTextControl *name;
public:
	SaveWindow(BRect rect, char *text);
	~SaveWindow() { printf("SaveWindow destructor\n"); fflush(stdout); }
	void MessageReceived(BMessage *message);
};

SaveWindow::SaveWindow(BRect rect, char *text) :
	BWindow(rect, "Save Field As", 
		B_TITLED_WINDOW_LOOK, 
		B_MODAL_APP_WINDOW_FEEL, 
		B_NOT_RESIZABLE | B_NOT_ZOOMABLE) {
	
	name = new BTextControl(BRect(0,0,rect.Width()*0.7, rect.Height()),
		"", "Name:", text, NULL);
	name->SetDivider(5*8); // depends on font size
	AddChild(name);
	ResizeTo(rect.Width(), name->Bounds().Height());
	rect = Bounds();
	name->MoveBy(0, (rect.Height()-name->Bounds().Height())/2.0);
	BButton *save = new BButton(BRect(rect.Width()*0.75,0,rect.Width(),rect.Height()), 
		"", "Save", new BMessage('Save'));
	AddChild(save);
	Show();
	printf("Show\n"); fflush(stdout);
}

void SaveWindow::MessageReceived(BMessage *message) {
	switch(message->what) {
	case 'Save': {
		BMessage m(MSG_SAVE);
		m.AddString("name", name->Text());
		my_app->window->PostMessage(&m);	
		Quit();
		break; }
	default:
		BHandler::MessageReceived(message);
	}
}

void AppWindow::Save() {
BRect rect(0, 0, 200, 50); rect.OffsetTo(view->ConvertToScreen(BPoint(20,20)));
BMenuItem *item = field->FindMarked();
char *text = (item == NULL) ? "Unnamed" : (char*)item->Label();
	new SaveWindow(rect, text);
}

void AppWindow::UpdateUnRedoMenu() {
BMenuItem *item = menubar->FindItem(MENU_GAME_UNDO);
	item->SetEnabled(view->moveList.CanUndo());
	item = menubar->FindItem(MENU_GAME_REDO);
	item->SetEnabled(view->moveList.CanRedo());
}

void AppWindow::MessageReceived(BMessage *message) {
BMenuItem *item;
int32 index;
BMessage *archive;
bool marked;
	switch(message->what) {
	case MENU_APP_NEW: 
		item = field->FindMarked();
		if (item != NULL) {
			view->SetBoard(fieldList.ItemAt(item->Message()->FindInt32("index")));
		}
		UpdateUnRedoMenu();
		break; 
	case MENU_GAME_UNDO:
		view->Undo();
		UpdateUnRedoMenu();
		break;
	case MENU_GAME_REDO:
		view->Redo();
		UpdateUnRedoMenu();
		break;
	case MENU_GAME_EDIT:
		item = menubar->FindItem(MENU_GAME_EDIT);
		marked = !item->IsMarked();
		item->SetMarked(marked);
		view->Edit(marked);
		menubar->FindItem(MENU_GAME_SAVE)->SetEnabled(marked);
		menubar->FindItem(MENU_GAME_DELETE)->SetEnabled(marked);
		UpdateUnRedoMenu();
		break;
	case MENU_GAME_SAVE:
		Save();
		break;
	case MENU_GAME_DELETE:
		item = field->FindMarked();
		if (item != NULL) {
			index = fieldList.Find(item->Label());
			if (index != -1) {
				fieldList.Remove(index);
				fieldList.BuildMenu(field, MENU_FIELD);
				SavePreferences();
			}
		}
		break;
	case MSG_SAVE:
		char *fieldText;
		message->FindString("name", (const char**)&fieldText);
		printf("fieldText: %s", fieldText); fflush(stdout);
		index = fieldList.Find(fieldText);
		if (index != -1) 
			fieldList.Replace(index, view->GetBoard());
		else {
			fieldList.Insert(fieldText, view->GetBoard());
			fieldList.BuildMenu(field, MENU_FIELD);
		}
		index = fieldList.Find(fieldText);
		item = field->FindItem(fieldText);
		item->SetMarked(true);
		SavePreferences();
		break;
	case MENU_FIELD:
		index = message->FindInt32("index");
		archive = fieldList.ItemAt(index);
		if (archive != NULL) view->SetBoard(new BMessage(archive));
		UpdateUnRedoMenu();
		break;
	case B_ABOUT_REQUESTED:
		AboutRequested();
		break;
	default:
		BWindow::MessageReceived(message);
	}
}


bool AppWindow::QuitRequested() {
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}

void AppWindow::AboutRequested() {
	BAlert *about = new BAlert("Solitaire", 
		"Solitaire.\nThis game is freeware.\n\n"
		"Written 1999.\n\n"
		"By Michael Pfeiffer.\n\n"
		"EMail: m.pfeiffer@jk.uni-linz.ac.at [expires 1.7.2000].","Close");
		about->Go();
}

App::App() : BApplication("application/x-vnd.solitaire") {
	BRect aRect;
	// set up a rectangle and instantiate a new window
	aRect.Set(100, 80, 100+500, 80+500);
	window = NULL;
	window = new AppWindow(aRect);		
}

static const char settingsFilename[] = "Solitaire settings";

void AppWindow::SavePreferences() {
BFile file;
BPath path;
	if (find_directory(B_COMMON_SETTINGS_DIRECTORY, &path) == B_OK) {
		path.Append(settingsFilename);
		if (file.SetTo(path.Path(), B_WRITE_ONLY | B_CREATE_FILE | B_ERASE_FILE) == B_OK) { 
			fieldList.Save(&file);
		}
	}
}

void AppWindow::RestorePreferences() {
BFile file;
BPath path;
	if (find_directory(B_COMMON_SETTINGS_DIRECTORY, &path) == B_OK) {
		path.Append(settingsFilename);
		if (file.SetTo(path.Path(), B_READ_ONLY) == B_OK) {
			fieldList.Restore(&file);
			return;
		} else { // load default fields
			BPath path(appPath);
			path.Append("Fields.data");
			if (file.SetTo(path.Path(), B_READ_ONLY) == B_OK) 
				fieldList.Restore(&file);
		} 
	}
}

int main(int argc, char *argv[]) { 
	be_app = NULL;
	new App();
	//my_app->window->view->Init();
	be_app->Run();
	delete be_app; 
	return 0;
}