#include "Application.h"
#include <TranslationUtils.h>
#include <stdio.h>


#define BUTTON_SIZE 0.19
bool xor1(bool a, bool b);

bool xor1(bool a, bool b) {
	return (a && !b) || (!a && b);
}

Board::Board(void) {
	Clear();
	int x, y;
	for (y = 0; y < 7; y++)
		for (x = 0; x < 7; x++) field[y][x] = true;
	field[3][3] = false;
	amount = 3*3-1+4*3*2;
}

Board::Board(BMessage *archive) {
bool *f;
ssize_t numBytes;
	if (B_OK == archive->FindData("field", (type_code)B_BOOL_TYPE, (const void**)&f, &numBytes)) {
		memcpy(field, f, sizeof(field));
		archive->FindInt16("amount", 0, &amount);
	} else
		Clear();
}

status_t Board::Archive(BMessage *archive, bool deep) {
	archive->AddData("field", B_BOOL_TYPE, field, sizeof(field));
	archive->AddInt16("amount", amount);
	return B_OK;
}

BArchivable *Board::Instantiate(BMessage *archive) {
	if (!validate_instantiation(archive, "Board"))
		return NULL;
	return new Board(archive);
}

bool Board::Get(int x, int y) {
	if (ValidPosition(x, y))
		return field[y][x];
	else
		return false;
}

void Board::Set(int x, int y, bool b) {
	if (ValidPosition(x, y)) {
		if (field[y][x] != b) {
			field[y][x] = b;
			amount = b ? amount + 1 : amount - 1;
		}
	}
}

void Board::Toggle(int x, int y) {
	Set(x, y, !Get(x, y));
}

void Board::Clear(void) {
int x, y;
	for (y = 0; y < 7; y++)
		for (x = 0; x < 7; x++) field[y][x] = false;
	amount = 0;
}

bool Board::ValidPosition(int x, int y) {
	if ((x < 0) || (x >= 7) || (y < 0) || (y >= 7)) return false;
bool a = x < 2, b = x >= 5, c = y < 2, d = y >= 5;
	return !(a&&c || a&&d || b&&c || b&&d); 
}

bool Board::Finish(void) {
int x, y;
	for (y = 0; y < 7; y++)
		for (x = 0; x < 7; x++)
			if (ValidPosition(x, y) && field[y][x] &&
				(ValidPosition(x-1, y) && ValidPosition(x+1,y) &&
					xor1(field[y][x-1], field[y][x+1]) ||
				 ValidPosition(x, y-1) && ValidPosition(x,y+1) &&
					xor1(field[y-1][x], field[y+1][x]) /*||
				 ValidPosition(x-1, y-1) && ValidPosition(x+1,y+1) &&
					xor1(field[y-1][x-1], field[y+1][x+1]) ||
				 ValidPosition(x-1, y+1) && ValidPosition(x+1,y-1) &&
					xor1(field[y+1][x-1], field[y-1][x+1])*/))
				return false;

	return true;
}

int Board::Amount(void) {
	return amount;
}

// ---------------------------------------------------------------------------

MoveList::MoveList() {
	board = NULL; top = current = 0;
}

void MoveList::Reset(Board *board) {
	this->board = board;
	top = current = 0;
}

bool MoveList::Redo(Move &move) {
	if (CanRedo()) {
		move = list[current++];
		Alternate(move);
		return true;
	} else 
		return false;
}

bool MoveList::Undo(Move &move) {
	if (CanUndo()) {
		move = list[--current];
		Alternate(move);
		return true;
	} else 
		return false;
}

void MoveList::Record(Coord from, Coord to) {
Move move(from, to);
	Alternate(move);
	list[current++] = move;
	top = current;
}

void MoveList::Alternate(Move &move) {
Coord coord((move.from.x+move.to.x)/2, (move.from.y+move.to.y)/2);
	board->Toggle(move.from.x, move.from.y);
	board->Toggle(move.to.x, move.to.y);
	board->Toggle(coord.x, coord.y);	
}

bool MoveList::CanUndo() {
	return current != 0;
}

bool MoveList::CanRedo() {
	return current < top;
}

// ---------------------------------------------------------------------------
BBitmap* ConvertBitmap(BBitmap* src,
         BRect dst,
         color_space color,
         bool preserve_aspect_ratio);

BBitmap* ConvertBitmap(BBitmap* src,
         BRect dst,
         color_space color,
         bool preserve_aspect_ratio) {
  // convert a bitmap from one size/color_space into a different
  // size color_space (possibly maintaining the aspect ratio) by
  // using the app_servers DrawBitmap

  // create a temporary bitmap and attach a view to it so we
  // can draw into it

  BBitmap* tmp = new BBitmap(dst, color, true);
  BBitmap* final = new BBitmap(dst, color);
  BView* view = new BView(dst, "", B_FOLLOW_ALL, B_WILL_DRAW);

  tmp->AddChild(view);
  tmp->Lock();
  view->SetHighColor(B_TRANSPARENT_32_BIT);
  view->FillRect(dst);

  if (preserve_aspect_ratio) {
      float f;
      BRect bounds(src->Bounds());

      if (bounds.Width() > bounds.Height()) {
    // leave width, adjust height to maintain aspect ratio
    f = bounds.Height() / bounds.Width() * dst.Height();
    dst.top = dst.top + ((dst.Height() - f) / 2);
    dst.bottom = dst.top + f;
      } else {
    // leave height, adjust width to maintain aspect ratio
    f = bounds.Width() / bounds.Height() * dst.Width();
    dst.left = dst.left + ((dst.Width() - f) / 2);
    dst.right = dst.left + f;
      }
  }

  view->DrawBitmap(src, src->Bounds(), dst);
  // we can draw do additional drawing at this point such as
  // placing a frame around the image or a 'bug' in the corner
  view->Sync();
  tmp->Unlock();
  // copy bits from temporary bitmap into final bitmap
  memcpy(final->Bits(), tmp->Bits(), tmp->BitsLength());
  // delete the temporary bitmap (deletes the attached view as well)
  delete tmp;
  return final;
}

BBitmap* Map(BBitmap *bmp);

BBitmap* Map(BBitmap *bmp) {
char *ptr = (char*)bmp->Bits();
uint32 *pixel, color;
int x, y;
BRect rect(bmp->Bounds());
	color = *(uint32*)ptr;
	for (y = 0; y <= rect.Height(); y++) {
		pixel = (uint32*)&ptr[y*bmp->BytesPerRow()];
		for (x = 0; x <= rect.Width(); x++, pixel++) {
			if (*pixel == color) 
				*pixel = B_TRANSPARENT_MAGIC_RGBA32;
		}
	}
	return bmp;
}

View::View(BRect rect) : BView(rect, NULL, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_SUBPIXEL_PRECISE | B_FRAME_EVENTS) {
	SetViewColor(B_TRANSPARENT_COLOR);
	board = new Board(); mouseMove = false;
	moveList.Reset(board);
	BBitmap *bmp;
	bmp = BTranslationUtils::GetBitmap("graphics/ball.tga");
	if (bmp == NULL) {
		fprintf(stderr, "solitaire error: could not open bitmap file\n");
		exit(-1);
	}
	bmPinTemplate = Map(ConvertBitmap(bmp, bmp->Bounds(), B_RGBA32, true));
	bmp = BTranslationUtils::GetBitmap("graphics/hollow.tga");	
	if (bmp == NULL) {
		fprintf(stderr, "solitaire error: could not open bitmap file\n");
		exit(-1);
	}
	bmHoleTemplate = Map(ConvertBitmap(bmp, bmp->Bounds(), B_RGBA32, true));
	mode = PLAY; bmPin = bmHole = NULL; 
	buffer = new BBitmap(rect, B_RGB16, true);
	FrameResized(rect.Width(), rect.Height());
}

void View::FrameResized(float width, float height) {
	if (width > height) {
		this->width = height;
		origin.Set((width-height)/2.0, 0);
	} else {
		this->width = width;
		origin.Set(0, (height-width)/2.0);
	}

	int w = (int)(this->width/7.0*2.0*BUTTON_SIZE);
	BRect rect(0, 0, w, w);
	
	if (bmPin != NULL) delete bmPin;
	if (bmHole != NULL) delete bmHole;

	bmPin = ConvertBitmap(bmPinTemplate, rect, B_RGB32, true);
	bmHole = ConvertBitmap(bmHoleTemplate, rect, B_RGB32, true);

	if ((buffer != NULL) && (buffer->Bounds() != Bounds())) {
		delete buffer;
		BRect rect(0, 0, width, height); 
		buffer = new BBitmap(rect, B_RGB16, true);
		DrawBuffer(rect);
	} 	
}

void View::DrawBuffer(BRect updateRect) {
BView view(buffer->Bounds(), NULL, B_FOLLOW_NONE, B_WILL_DRAW);
BRect rect = Bounds(); 
int x, y;
float wx, wy, d = width/7.0, d2;
	wy = d / 2.0; d2 = d*BUTTON_SIZE; 
	wy -= d2;
float x0 = origin.x , x1 = x0 + width/3.0-d/2.0, 
	  x2 = x0 + 2.0*width/3.0+d/2.0, x3 = x0 + width,
	  y0 = origin.y, y1 = y0+width/3.0 - wy, 
	  y2 = y0 + 2.0*width/3.0+wy, y3 = y0 + width;

BPoint polygon[12];
#define CROSS
#ifdef CROSS
	polygon[0] = BPoint(x1, y0);	
	polygon[1] = BPoint(x2, y0);	
	polygon[2] = BPoint(x2, y1);	
	polygon[3] = BPoint(x3, y1);	
	polygon[4] = BPoint(x3, y2);	
	polygon[5] = BPoint(x2, y2);	
	polygon[6] = BPoint(x2, y3);	
	polygon[7] = BPoint(x1, y3);	
	polygon[8] = BPoint(x1, y2);	
	polygon[9] = BPoint(x0, y2);	
	polygon[10] = BPoint(x0, y1);	
	polygon[11] = BPoint(x1, y1);	
#else
	polygon[0] = BPoint(x1, y0);	
	polygon[1] = BPoint(x2, y0);	
	polygon[2] = BPoint(x3, y1);	
	polygon[3] = BPoint(x3, y2);	
	polygon[4] = BPoint(x2, y3);	
	polygon[5] = BPoint(x1, y3);	
	polygon[6] = BPoint(x0, y2);	
	polygon[7] = BPoint(x0, y1);	
#endif
	buffer->Lock();	
	buffer->AddChild(&view);

	//view.SetHighColor(128,128,255);
	view.SetHighColor(255,255,255);
	view.FillRect(Bounds());

	//view.SetHighColor(0,0,255);
	#define COLOR 176
	view.SetHighColor(COLOR,COLOR,COLOR);
#ifdef CROSS
	view.FillPolygon(polygon, 12);
	view.SetHighColor(0,0,0);
	view.StrokePolygon(polygon, 12);
#else
	view.FillPolygon(polygon, 8);
	view.SetHighColor(0,0,0);
	view.StrokePolygon(polygon, 8);
#endif
	view.SetDrawingMode(B_OP_OVER);
	wy += y0;
	for (y = 0; y <= 6; y++, wy+=d)
		for (wx = x0+d/2.0-d2, x = 0; x <= 6; x++, wx+=d) 
			if (board->ValidPosition(x, y)) {
				rect = BRect(wx-d2, wy-d2, wx+d2, wy+d2);
				if (board->Get(x, y)) { // draw pin
					view.DrawBitmap(bmPin, BPoint(wx, wy));
					//FillEllipse(rect);
				} else { // draw hole
					view.DrawBitmap(bmHole, BPoint(wx, wy));
					//StrokeEllipse(rect);
				}
			} //else StrokeRect(BRect(wx-dx2, wy-dy2, wx+dx2, wy+dy2));

	view.Sync();
	
	buffer->RemoveChild(&view);
	buffer->Unlock();	
}


void View::Draw(BRect updateRect) {
	if ((buffer != NULL) && (buffer->Bounds() == Bounds())) {
		DrawBitmap(buffer, updateRect, updateRect);
	}
}

void View::Position(BPoint point, int &x, int &y) {
/*BRect rect = Bounds();
	x = (int)(7.0*point.x/(rect.Width()+1));
	y = (int)(7.0*point.y/(rect.Height()+1));
*/
	if (point.x < origin.x || point.y < origin.y || point.x-origin.x > width
		|| point.y-origin.y > width) {
		x = y = 0;
	} else {
		x = (int)(7.0*(point.x-origin.x)/(width+1));
		y = (int)(7.0*(point.y-origin.y)/(width+1));
	}
}

BRect View::RectFromPosition(int x, int y) {
float w = width/7.0,
	left = origin.x+w*x, top = origin.y+w*y;
	return BRect(left, top, left+w, top+w);
}

void View::MouseDown(BPoint point) {
int x, y;
	Position(point, x, y);
	SetMouseEventMask(B_POINTER_EVENTS);
	if (mode == PLAY && board->ValidPosition(x, y) && board->Get(x, y)) {
		//if (mouseMove) board->Set(position.x, position.y, true);
		//else 
		mouseMove = true; 
		position.x = x; position.y = y;
		BRect rect = bmPin->Bounds();
		board->Set(x, y, false); 
		DrawBuffer(RectFromPosition(x, y));
		Invalidate(RectFromPosition(x, y));
		BBitmap * bmp = new BBitmap(rect, bmPin->ColorSpace(), false);
		memcpy(bmp->Bits(), bmPin->Bits(), bmPin->BitsLength());
		DragMessage(new BMessage(), bmp, B_OP_OVER, BPoint(rect.Width()/2, rect.Height()/2));
		//DragMessage(new BMessage(), bmp, BPoint(8, 8));
	} else if (mode == EDIT && board->ValidPosition(x, y)) {
		board->Set(x, y, !board->Get(x, y));
		DrawBuffer(RectFromPosition(x, y));
		Invalidate(RectFromPosition(x, y));
	}
}

bool View::ValidMove(int x, int y) {
	return board->ValidPosition(x, y) && 
		 xor1(abs(x-position.x) == 2, abs(y-position.y) == 2) &&  // vertical or horizontal only 
		// (abs(x-position.x) == 2 || abs(y-position.y) == 2) && 
		!board->Get(x, y) && 
		board->Get((x+position.x)/2, (y+position.y)/2);
}

void View::MouseMoved(BPoint point, uint32 transit, const BMessage *message) {
}

void View::MouseUp(BPoint point) {
int x, y;
Coord from, to;
	if (!mouseMove) return;
	Position(point, x, y);
	mouseMove = false;
	if (ValidMove(x, y)) {
		//BRect rect = RectFromPosition(x, y);
		from = Coord(x, y);
		to = Coord(position.x, position.y);
		board->Set(position.x, position.y, true);
		moveList.Record(from, to);
		((AppWindow*)Window())->UpdateUnRedoMenu();
		//board->Set(x, y, true);
		//board->Set((x+position.x)/2, (y+position.y)/2, false);
		//rect = rect | RectFromPosition(position.x, position.y);
		UpdateRect(Move(from, to));
		//DrawBuffer(rect);
		//Invalidate(rect);
		if (board->Finish()) {
			char string[80];
			sprintf(string, "Game over!\n%d pieces left.", board->Amount());
			BAlert *finish = new BAlert("Finish", string, "Close");
			finish->Go();			
		}
	} else {
		board->Set(position.x, position.y, true);
		DrawBuffer(RectFromPosition(position.x, position.y));
		Invalidate(RectFromPosition(position.x, position.y));
	}
}

BMessage *View::GetBoard() {
BMessage *msg = new BMessage();
	if (B_OK == board->Archive(msg, false))
		return msg;
	else {
		delete msg; return NULL;
	}
}

void View::SetBoard(BMessage *archive) {
	delete board;
	board = new Board(archive);
	moveList.Reset(board);
	DrawBuffer(Bounds());
	Invalidate();
}

void View::Edit(bool edit) {
	if (edit) mode = EDIT;
	else mode = board->Finish() ? STOP : PLAY;
	moveList.Reset(board);
}

void View::UpdateRect(const Move &move) {
BRect rect = RectFromPosition(move.from.x, move.from.y);
	rect = rect | RectFromPosition(move.to.x, move.to.y);
	DrawBuffer(rect);
	Invalidate(rect);
}

void View::Undo() {
Move move;
	if (moveList.Undo(move)) UpdateRect(move);
}

void View::Redo() {
Move move;
	if (moveList.Redo(move)) UpdateRect(move);
}