#ifndef VIEW_H
#define VIEW_H

class _EXPORT Board : public BArchivable {
	bool field[7][7];
	int16 amount;
public:
	Board(void);
	Board(BMessage *archive);
	status_t Archive(BMessage *archive, bool deep);
	static BArchivable *Instantiate(BMessage *archive);
	
	void Clear(void);
	bool Get(int x, int y);
	void Set(int x, int y, bool b);
	void Toggle(int x, int y);
	bool ValidPosition(int x, int y);
	bool Finish();
	int Amount();
};

class Coord{
public:
	uint8 x, y;
	Coord(uint8 x, uint8 y) : x(x), y(y) {};
	Coord(Coord &coord) : x(coord.x), y(coord.y) {};
	Coord() {};
};

class Move {
public:
	Coord from, to;
	Move() {};
	Move(Coord from, Coord to) : from(from), to(to) {};
};

class MoveList {
	Board *board;
	Move list[7*7];
	int top, current;
	
	void Alternate(Move &move);
public:
	MoveList();
	void Reset(Board *board);
	bool Redo(Move &move);
	bool Undo(Move &move);
	void Record(Coord from, Coord to);
	inline bool CanUndo();
	inline bool CanRedo();
};
 
class View : public BView {
	Board *board;
	Coord position;
	bool mouseMove;
	
	BPoint origin; // of board square
	float width;   
	
	BBitmap *bmPin, *bmHole;
	BBitmap *bmPinTemplate, *bmHoleTemplate;
	enum {PLAY, EDIT, STOP} mode;
	
	BBitmap *buffer;
public:
	MoveList moveList;

	View(BRect rect);
	~View() { delete bmPin; }
	void Draw(BRect updateRect);
	void DrawBuffer(BRect rect);
	void FrameResized(float width, float height);
	void MouseDown(BPoint point);
	void MouseUp(BPoint point);
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message);
	void Position(BPoint point, int &x, int &y);
	bool ValidMove(int x, int y);
	BMessage *GetBoard();
	void SetBoard(BMessage *archive);
	void Edit(bool edit);
	BRect RectFromPosition(int x, int y);
	
	void Undo();
	void Redo();
	void UpdateRect(const Move &move);
};
#endif
